#!/usr/bin/perl
#
# Author  : Yann Stettler
# Version : 1.0
#
# Desc :
# Continuously ping several hosts and return the results when asked.
#
# Accept incoming TCP connections on $PORT. The first line must contain the
# secret. The second line should be a host/IP address or the keyword "summary".
#
# - If the 2nd line is a host/IP that is monitored, return the collected data.
# - If the 2nd line is a host/IP that isn't monitored, add it to the list of
#   hosts to ping
# - If the 2nd line is "summary", return data for all the hosts monitored
#
# Arguments:
#
# -f = run in foreground (don't detach)
#
# Packages needed:
# IO::Socket
# IO::Socket::Timeout
# PDL
# IPC::Shareable
# Data::Validate::Domain
# Data::Validate::IP 
# Proc::Daemon
#
use strict;
use warnings;
use POSIX qw(strftime);
use Time::HiRes qw(gettimeofday sleep);
use IO::Socket;
use IO::Socket::Timeout;
use Errno qw(ETIMEDOUT EWOULDBLOCK);
use PDL;
use IPC::Shareable;
use Data::Validate::Domain qw(is_domain);
use Data::Validate::IP qw(is_ipv4 is_ipv6);
use Proc::Daemon;

my $fg=shift;
if (!$fg || $fg ne '-f') {
   Proc::Daemon::Init();
}

#############################################################################
# Configuration                                                             #
#############################################################################
my $FPING="/usr/local/sbin/fping";
my $INTERVAL=60*5;                  # Must match cacti poller interval (in sec)
my $MARGIN=5;                       # Margin (in sec) between loops
my $FREQ=200;                       # Time, in ms, between two pings
my $BINDTO='0.0.0.0';               # On which address we listen
my $PORT=9928;                      # On which port we listen for connections
my $SECRET="12345";                 # Must match secret in cacti script
my $LOG=0;

# Initial list of hosts to ping. Will learn new hosts based on requests
# my @HostList=("192.168.1.10", "192.168.100.1");
my @HostList;

#############################################################################

my $nbping=int($INTERVAL/$FREQ*1000/5);  # Do a fast initial run

if (! -x $FPING) {
   die "Can't execute $FPING: $!\n";
}

my $strResults;
my $strHostList;
my $child;

my $handleout = (tie $strResults, 'IPC::Shareable', undef, { mode => 0600, destroy => 1 }) or die "parent: tie failed\n";
my $handlein = (tie $strHostList, 'IPC::Shareable', undef, { mode => 0600, destroy => 1 }) or die "parent: tie failed\n";

$strResults="";
$strHostList=join(':', @HostList);

unless ($child = fork) {        # i'm the child
    die "cannot fork: $!" unless defined $child;
    while(1) {
       my $start=gettimeofday();
       if ($strHostList eq "") {
          # Nothing to ping, better sleep a little
          sleep(10);
       } else {
          if (!$nbping) { $nbping=1; }
          &log("Doing $nbping pings with a fequency of $FREQ ms to $strHostList");
          &doping();

          # Heartbeat
          my $elapsed=gettimeofday() - $start;
          my $avgptime=$elapsed / $nbping;
          &log(sprintf("Sent $nbping pings in %d sec (target: $INTERVAL sec)", $elapsed));
          if ($elapsed < $INTERVAL) {
             if ($elapsed < ($INTERVAL-$MARGIN)) {
                $nbping += int(($INTERVAL-$elapsed-$MARGIN)/$avgptime);
             }
             sleep($INTERVAL-$elapsed);
          } elsif ($elapsed > $INTERVAL) {
             $nbping -= int(($elapsed-$INTERVAL+$MARGIN)/$avgptime);
          }
       }
    }
    exit 0;
} 

$SIG{'INT'}='stop_deamon';
$SIG{'TERM'}='stop_deamon';
$SIG{'HUP'}='stop_deamon';

my $socket = new IO::Socket::INET (
    LocalHost => $BINDTO,
    LocalPort => $PORT,
    Proto => 'tcp',
    Listen => 5,
    Reuse => 1
) or die "cannot create socket $!\n";

&log("Accepting connection on port $PORT");

$| = 1;
while (1) {

    my $client_socket = $socket->accept();
    IO::Socket::Timeout->enable_timeouts_on($client_socket);
    $client_socket->read_timeout(3);
    $client_socket->write_timeout(3);

    my $client_address = $client_socket->peerhost();

    my $ligndata="";
    my $lign=$client_socket->getline;
    if (! $lign || (! $lign && ( 0+$! == ETIMEDOUT || 0+$! == EWOULDBLOCK ))) {
       shutdown($client_socket, 1);
       next;
    }

    $lign =~ s/\s+$//;
    if ($lign eq $SECRET) {
       $ligndata=$client_socket->getline;
       if (! $ligndata || (! $ligndata && ( 0+$! == ETIMEDOUT || 0+$! == EWOULDBLOCK ))) {
          shutdown($client_socket, 1);
          next;
       }
       $ligndata=~s/\s+$//;
       &log("Connection from $client_address : $ligndata");
       if ($ligndata ne "" && (is_domain($ligndata) || is_ipv4($ligndata))) {
          my $found=0;
          foreach my $i (0 .. $#HostList) {
             if ($HostList[$i] eq $ligndata) {
                $found=1;
                last;
             }
          }
          if (!$found) { 
             push(@HostList, $ligndata);
             $strHostList=join(':', @HostList);
          }
          $client_socket->send(&get_data($ligndata));
       } elsif ($ligndata eq "summary") {
          $client_socket->send(&get_data(""));
       }
    } else {
       &log("Access denied for $client_address");
    }
    shutdown($client_socket, 1);
}
$socket->close();

$handleout->remove;
$handleout->clean_up_all;
$handlein->remove;
$handlein->clean_up_all;

exit 0;


sub get_data {
   my ($host) = @_;

   my @HostData=split(/\n/, $strResults);
   my $buf="";
   if ($host eq "") {
      $buf = sprintf("%-28s %5s %5s %5s %7s %7s %7s %7s\n", "Host", "Sent", "Lost", "%", "Min", "Max", "Avg", "Dev");
      $buf .= "------------------------------------------------------------------------------\n";
   }

   foreach my $i (0 .. $#HostData) {
      my @Results=split(/:/, $HostData[$i]);
      if ($host eq "") {
         my $lost = $Results[2]*100/$Results[1];
         $buf .= sprintf("%-28s %5d %5d %5.1f %7.3f %7.3f %7.3f %7.3f\n", $Results[0], $Results[1], $Results[2], $lost, $Results[3], $Results[4], $Results[5], $Results[6]);
      } elsif ($Results[0] eq $host) {
         if ($Results[1] == $Results[2]) {
            $buf = sprintf("%s min:U avg:U max:U dev:U loss:100.00", $Results[0]);
         } else {
            my $lost = $Results[2]*100/$Results[1];
            $buf = sprintf("%s min:%0.4f avg:%0.4f max:%0.4f dev:%0.4f loss:%0.4f", $Results[0], $Results[3], $Results[5], $Results[4], $Results[6], $lost);
         }
         last;
      }
   }
   $buf .= "\n";
   return($buf);
}

sub doping {
   my ($target, $tmp, $lost, $tmpbuf, @pingdata);
   my @HostList=split(/:/, $strHostList);

   open(FP, "$FPING -i 5 -t 100 -p $FREQ -C $nbping -q @HostList 2>&1 |") || die "Can't run $FPING: $!\n";
   while(<FP>) {
      chop;
      if (($target,$tmp) = /^(\S+)\s*:\s+(.*)$/) {
         $lost=$tmp=~tr/-//;
         $tmp=~s/-//g;
         if ($tmp=~/[^\d\.\s]+/) {
            &log("ERROR: $tmp");
         } else {
            @pingdata = split(/\s+/, $tmp);

            if ($#pingdata > -1) {
               my $piddle1 = pdl @pingdata;
               my ($mean,$prms,$median,$min,$max,$adev,$rms) = statsover $piddle1;

               my $total=$lost + scalar(@pingdata);
               $tmpbuf .= "$target:$total:$lost:$min:$max:$mean:$prms\n";
            }
         }
      }
   }
   close(FP);
   $strResults=$tmpbuf;
}

sub stop_deamon {

   kill(15, $child);

   $socket->close();
   $handleout->remove;
   $handleout->clean_up_all;
#   $handlein->remove;
   $handlein->clean_up_all;
   exit 0;
};

sub log {
   my ($msg) = @_;

   if ($LOG) {
      print strftime '%d.%m.%Y %H:%M:%S - ', localtime(time);
      print $msg . "\n";
   }
}
